/*********************************************************************
 *              Copyright (c) 2021 PikeTec GmbH
 *               CONFIDENTIAL AND PROPRIETARY.
 *                   ALL RIGHTS RESERVED
 *              - PROPRIETARY OF PikeTec GmbH -
 **********************************************************************/
/* clang-format off */

#ifndef __TPT_FUSION_H_
#define __TPT_FUSION_H_

#include "tpt_fusion_base.h"
#include <stdarg.h>

#ifdef __cplusplus
extern "C"
{
#endif

/* =============================== DECLARING VARIABLES ========================================= */

/**
 * Declares a node's scalar variable on the FUSION and returns a handle that allows 
 * to read and write that variable during a FUSION session. 
 *
 * In case, a variable with the given name has already been declared, it will be 
 * returned, if its type and role match the declaration, otherwise, an error is 
 * set. 
 *
 * @param handle
 *          The handle of the calling node.
 * @param name
 *          The name of the scalar variable to declare. 
 * @param role
 *          The role of the scalar variable to declare.
 *
 * @return A handle to the variable for accessing it during the FUSION session.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_declareScalar(tpt_fusion_node_handle handle,
                                                    const char *name,
                                                    tpt_fusion_variable_role role);

/**
 * Declares a node's string variable on the FUSION and returns a handle that allows 
 * to read and write that string variable during a FUSION session.  The strings capacity 
 * (maximum length incl. terminating '\0') is not part of the declaration. Instead, 
 * it will be set each time, the underlying array is written. Therefore the length 
 * of a string may vary during a FUSION session.
 *
 * In case, a variable with the given name has already been declared, it will be 
 * returned, if its type and role match the declaration, otherwise, an error will 
 * be set.
 *
 * @param handle
 *          The handle of the calling node.
 * @param name 
 *          The name of the string to declare.
 * @param role
 *          The role of the string to declare.
 *
 * @return A handle to the variable for accessing it during the FUSION session.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_declareString(tpt_fusion_node_handle handle,
                                                    const char *name,
                                                    tpt_fusion_variable_role role);

/**
 * Declares a node's array variable on the FUSION and returns a handle that allows 
 * to read and write that array variable during a FUSION session.  The arrays length 
 * is not part of the declaration. Instead, it will be determined when the array 
 * is written for the first time using tpt_fusion_writeValue() or tpt_fusion_writeValueAsLong().
 *
 * <b>Compatibility notice:</b>
 * In some cases arrays are used to represent anonymous axes of curves or maps 
 * (see tpt_fusion_declareCurve() and tpt_fusion_declareMap() for details). 
 * In older versions of TPT (before 5.0) the anonymous axes naming convention has been 
 * "<curvename>:XAXIS"/"<mapname>:XAXIS"/"<mapname>:YAXIS", whereas in new versions 
 * (since TPT5.0) it is "<curcename>.xaxis"/"<mapname>.xaxis"/"<mapname>.yaxis". 
 * If a node implementing the outdated naming scheme, declares an array as, e.g.,
 * "myaxis:XAXIS" the name will be silently replaced by "myaxis.xaxis" to ensure
 * consistency of all declarations between different nodes!
 *
 * In case, a variable with the given name has already been declared, it will be 
 * returned, if its type and role match the declaration, otherwise, an error will 
 * be set.
 *
 * @param handle
 *          The handle of the calling node.
 * @param name 
 *          The name of the array to declare.
 * @param role
 *          The role of the array to declare.
 *
 * @return A handle to the variable for accessing it during the FUSION session.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_declareArray(tpt_fusion_node_handle handle,
                                                   const char *name,
                                                   tpt_fusion_variable_role role);

/**
 * Declares a node's curve variable on the FUSION and returns a handle that allows 
 * to read and write that variable during a FUSION session. 
 *
 * In order to define a curve's x-axis, an array variable must be provided. If this 
 * x-axis array is anonymous (has no name), a new array variable must be declared 
 * using the naming scheme "<curvename>.xaxis" (e.g. "MYCURVE.xaxis"). 
 * This convention ensures, that all nodes use the same anonymous x-axis for the 
 * same curve.
 *
 * <b>Compatibility notice:</b>
 * In older versions of TPT (before 5.0) the anonymous axis convention has been 
 * "<curvename>:XAXIS". If a node implementing this outdated naming scheme,  
 * declares such an axis name it will be silently replaced by "<curvename>.xaxis" 
 * to ensure consistency of all declarations between different nodes!
 *
 * In case, a variable with the given name has already been declared, it will be 
 * returned, if its type, role and x-axis match the declaration, otherwise, an 
 * error will be set.
 *
 * @param handle
 *          The handle of the calling node.
 * @param name 
 *          The name of the curve to declare.
 * @param xaxis
 *          The x-axis of the curve to declare.
 * @param role
 *          The role of the curve to declare. 
 *
 * @return A handle to the variable for accessing it during the FUSION session.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_declareCurve(tpt_fusion_node_handle handle,
                                                   const char *name,
                                                   tpt_fusion_variable_handle xaxis,
                                                   tpt_fusion_variable_role role);

/**
 * Declares a node's map variable on the FUSION and returns a handle that allows 
 * to read and write that variable during a FUSION session. 
 *
 * In order to define a curve's x- or y-axis, an array variable must be provided. 
 * If this axis array is anonymous (has no name), a new array variable must be 
 * declared using the corresponding naming scheme 
 *
 *   - "<mapname>.xaxis" (e.g. "MYMAP.xaxis"). 
 *   - "<mapname>.yaxis" (e.g. "MYMAP.yaxis"). 
 *
 * This ensures, that all nodes use the same anonymous x-axis/y-axis for the same map.
 *
 * <b>Compatibility notice:</b>
 * In older versions of TPT (before 5.0) the anonymous axes convention has been 
 * "<mapname>:XAXIS"/"<mapname>:YAXIS". If a node implementing this outdated 
 * naming scheme, declares such an axis name it will be silently replaced by 
 * "<mapname>.xaxis"/"<mapname>.yaxis" to ensure consistency of all declarations
 * between different nodes!
 *
 * In case, a variable with the given name has already been declared, it will be 
 * returned, if its type, role and its axes match the declaration, otherwise, an 
 * error will be set.
 *
 * @param handle
 *          The handle of the calling node.
 * @param name 
 *          The name of the map to declare.
 * @param xaxis
 *          The x-axis of the map to declare.
 * @param yaxis
 *          The y-axis of the map to declare.
 * @param role
 *          The role of the map to declare. 
 *
 * @return A handle to the variable for accessing it during the FUSION session.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_declareMap(tpt_fusion_node_handle handle,
                                                 const char *name,
                                                 tpt_fusion_variable_handle xaxis,
                                                 tpt_fusion_variable_handle yaxis,
                                                 tpt_fusion_variable_role role);

/**
 * This function is called by nodes that should not affect the "declaration count" for
 * declared variables. See documentation of <code>tpt_fusion_declarationCount()</code> for
 * details.
 * Usually, there is no need or reason for custom nodes to call this functions. It is intended
 * for TPT service nodes basically. 
 */
TPT_DECLARE_FUSION(void) tpt_fusion_daemonNode(tpt_fusion_node_handle handle);

/**
 * For a given variable this function returns the total number of declareXxxx()-calls that have been
 * executed by one or several fusion nodes, except for those nodes that have been tagged as 
 * "daemon-nodes". The declaration count is used as an indicator of how many nodes
 * in total request access to a particular variable. Please note that if one of the nodes calls 
 * tpt_fusion_declareXxxx() several times (which is bad design anyway) this will be counted several
 * times.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read.
 *
 * @return total number of calls of tpt_fusion_declareXxxx(), which is always at minimum 1.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_declarationCount(tpt_fusion_node_handle handle, 
                                                 tpt_fusion_variable_handle vhandle);

/* =============================== RETRIEVING VARIABLES ======================================== */

/**
 * Retrieves the FUSION session's variable for the given name or returns zero/null 
 * if a variable with that name has not been declared. 
 *
 * @param handle
 *          The handle of the calling node.
 * @param name 
 *          The name of the variable to retrieve.
 *
 * @return The FUSION session's variable for the given name or returns zero/null 
 *         if a variable with that name has not been declared.
 */
TPT_DECLARE_FUSION(const tpt_fusion_variable_handle) tpt_fusion_getVariable(tpt_fusion_node_handle handle, 
                                                        const char *name);

/**
 * Returns an array of all variables that have been declared on the FUSION. The 
 * resulting array's length can be retrieved via 'tpt_fusion_getVariablesCount()'.
 *
 * NOTE: Callers must free the returned array with 'tpt_fusion_freeArray()'.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return An array of all variables that have been declared on the FUSION. 
 */
TPT_DECLARE_FUSION(const tpt_fusion_variable_handle *) tpt_fusion_getVariables(tpt_fusion_node_handle handle);

/**
 * Returns the total count of all variables declared on the FUSION.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return The total count of all variables declared on the FUSION.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getVariablesCount(tpt_fusion_node_handle handle);


/* =============================== READING VARIABLES ========================================= */

/**
 * Returns a read-only pointer to the current value of the given variable as a double value. 
 * If the given variable is a map or curve, a pointer to its value(s) is returned.
 *
 * If the given variable is an array, curve or map, the result is a pointer to an 
 * array with an arbitrary number of values. In such cases, one may either call 
 * 'tpt_fusion_getVariableLength()' with the variable in question to retrieve its 
 * length or - alternatively - use 'tpt_fusion_readValues()'.
 *
 * This function is actually a shortcut for the following call:
 *
 *   const double *v = tpt_fusion_readValues(handle, vhandle, 0)
 *
 * NOTE: In order to read an axis of a map or curve, one may retrieve the
 *       axis variable first with a call to tpt_fusion_getVariableXAxis() or 
 *       tpt_fusion_getVariableYAxis().
 * 
 * NOTE: Each variable can be read using this function even if the variable has been written
 *       as a long value. FUSION will automatically cast the values accordingly. However,
 *       int64 values bigger than 2^53 might cause imprecise value casts! Thus, it is
 *       recommended to read double values as double and long values as long to ensure data
 *       precision even for int64 values.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read. 
 *
 * @return A read-only pointer to the current value of the given variable. 
 */
TPT_DECLARE_FUSION(const double *) tpt_fusion_readValue(tpt_fusion_node_handle handle, 
                                   tpt_fusion_variable_handle vhandle);

/**
 * Returns a read-only pointer to the current value of the given variable as a long value. 
 * If the given variable is a map or curve, a pointer to its value(s) is returned.
 *
 * If the given variable is an array, curve or map, the result is a pointer to an 
 * array with an arbitrary number of values. In such cases, one may either call 
 * 'tpt_fusion_getVariableLength()' with the variable in question to retrieve its 
 * length or - alternatively - use 'tpt_fusion_readValuesAsLong()'.
 *
 * This function is actually a shortcut for the following call:
 *
 *   const tpt_long *v = tpt_fusion_readValuesAsLong(handle, vhandle, 0)
 *
 * NOTE: In order to read an axis of a map or  curve, one may retrieve the
 *       axis variable first with a call to tpt_fusion_getVariableXAxis() or 
 *       tpt_fusion_getVariableYAxis().
 * 
 * NOTE: Each variable can be read using this function even if the variable has been written
 *       as a double value. FUSION will automatically cast the values accordingly. However,
 *       reading double values as long will cause loss of precision and might cause truncation 
 *       for values bigger than 2^63-1 or smaller than -2^63! 
 *       Thus, it is recommended to read double values as double and long values as long to
 *       ensure data precision even for int64 values.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read. 
 *
 * @return A read-only pointer to the current value of the given variable. 
 */
TPT_DECLARE_FUSION(const tpt_long *) tpt_fusion_readValueAsLong(tpt_fusion_node_handle handle, 
                                           tpt_fusion_variable_handle vhandle);

/**
 * Returns a read-only pointer to the current value of the given variable as a double value. 
 * If the given variable is a map or curve, a pointer to its value(s) is returned.
 * If the argument 'length' is not null/zero, the total number of values (total dimension)
 * will be stored to that address.
 *
 * NOTE: In order to read an axis of a map or curve, one may retrieve the
 *       axis variable first with a call to tpt_fusion_getVariableXAxis() or 
 *       tpt_fusion_getVariableYAxis().
 * 
 * NOTE: Each variable can be read using this function even if the variable has been written
 *       as a long value. FUSION will automatically cast the values accordingly. However,
 *       int64 values bigger than 2^53 might cause imprecise value casts! Thus, it is
 *       recommended to read double values as double and long values as long to ensure data
 *       precision even for int64 values.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read. 
 * @param length 
 *          The pointer to an unsigned integer, that will store the values count (total
 *          dimension).
 *
 * @return A read-only pointer to the current value of the given variable. 
 */
TPT_DECLARE_FUSION(const double *) tpt_fusion_readValues(tpt_fusion_node_handle handle, 
                                    tpt_fusion_variable_handle vhandle,
                                    unsigned int *length);

/**
 * Returns a read-only pointer to the current value of the given variable as a long value. 
 * If the given variable is a map or curve, a pointer to its value(s) is returned.
 * If the argument 'length' is not null/zero, the total number of values (total dimension)
 * will be stored to that address.
 *
 * NOTE: In order to read an axis of a map or curve, one may retrieve the
 *       axis variable first with a call to tpt_fusion_getVariableXAxis() or 
 *       tpt_fusion_getVariableYAxis().
 * 
 * NOTE: Each variable can be read using this function even if the variable has been written
 *       as a double value. FUSION will automatically cast the values accordingly. However,
 *       reading double values as long will cause loss of precision and might cause truncation 
 *       for values bigger than 2^63-1 or smaller than -2^63! 
 *       Thus, it is recommended to read double values as double and long values as long to
 *       ensure data precision even for int64 values.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read. 
 * @param length 
 *          The pointer to an unsigned integer, that will store the values count (total
 *          dimension).
 *
 * @return A read-only pointer to the current value of the given variable. 
 */
TPT_DECLARE_FUSION(const tpt_long *) tpt_fusion_readValuesAsLong(tpt_fusion_node_handle handle, 
                                            tpt_fusion_variable_handle vhandle,
                                            unsigned int *length);

/**
 * This method must be called during <code>tpt_fusion_node_callfcn()</code> only. 
 * It returns true iff the variable <code>vhandle</code> has been updated during the most recent
 * FUSION cycle.
 * More precisely: it returns true iff between the most recent and the current call of 
 * <code>tpt_fusion_node_callfcn()</code> for the FUSION node <code>handle</code>, at least 
 * one <i>other</i> FUSION node has called one of the functions <code>tpt_fusion_write*()</code>
 * for the variable <code>vhandle</code>.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to check for update.
 * @return true iff the signal has been updated during the most recent FUSION cycle
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_isUpdated(tpt_fusion_node_handle handle,
                                           tpt_fusion_variable_handle vhandle);

/**
 * This method must be called during <code>tpt_fusion_node_callfcn()</code> only.
 * Calling this method is usually not required. It changes the flag whether or not the variable
 * <code>vhandle</code> has been updated during this call of <code>tpt_fusion_node_callfcn()</code> or not.
 * However, this information is tracked automatically when calling tpt_fusion_write*() 
 * functions anyway.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_setUpdated(tpt_fusion_node_handle handle,
                      tpt_fusion_variable_handle vhandle, unsigned int updated);

/**
 * returns true if the variable has been written as a long value (or an array of long values) 
 * recently. The function returns false if the value has been written as
 * double.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read. 
 *
 * @return true iff the variable has been written as long values instead of double values 
 *         recently
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_isWrittenAsLong(tpt_fusion_node_handle handle, 
                                           tpt_fusion_variable_handle vhandle);

/* =============================== LAZY HANDLING OF READING/WRITING VALUES ====================== */

/**
 * Usually the value of a variable has to be written before it can be read. However, for each
 * variable a so called lazy-handler can be installed that is responsible for providing the 
 * variable's value just before the value requested to be read.
 *
 * If a variable defines a lazy-handler already, the existing one will be replaced by the new one.
 * 
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param lazyhandler
 *          pointer to the handler function or 0 if the handler should be uninstalled. 
 */
TPT_DECLARE_FUSION(void) tpt_fusion_installLazyHandler(tpt_fusion_node_handle handle, 
                                   tpt_fusion_variable_handle vhandle, 
                                   tpt_fusion_lazy_handler lazyhandler);

/**
 * returns true if the variable has a lazy handler installed. On the other hand, the function 
 * returns false if the value has no lazy handler.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to read. 
 *
 * @return true iff the variable has an installed lazy handler
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_hasLazyHandler(tpt_fusion_node_handle handle, 
                                   tpt_fusion_variable_handle vhandle);

/* =============================== WRITING VARIABLES ========================================= */

/**
 * Writes a new scalar double value to the given variable on the FUSION. This 
 * function is actually a shortcut for the following call:
 *
 *  tpt_fusion_writeValues(handle, vhandle, value, 1);
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param value
 *          The pointer from which the variable's new value will be copied from. 
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeValue(tpt_fusion_node_handle handle, 
                           tpt_fusion_variable_handle vhandle, 
                           const double *value);

/**
 * Writes a new scalar long value to the given variable on the FUSION. This 
 * function is actually a shortcut for the following call:
 *
 *  tpt_fusion_writeValuesAsLong(handle, vhandle, values, 1);
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param value
 *          The pointer from which the variable's new value will be copied from. 
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeValueAsLong(tpt_fusion_node_handle handle, 
                                 tpt_fusion_variable_handle vhandle, 
                                 const tpt_long *value);

/**
 * Writes new double values to the given variable on the FUSION. If the given variable 
 * is a map or curve, this function will always write the values. The total 
 * number of values copied from the given value pointer depends on the given total "length"
 * of the values. If length > 1, the array is assumed to be 1-dimensional only. Use 
 * tpt_fusion_writeNDimValues() to write n-dimensional arrays.
 *
 * NOTE: In order to write an axis of a map or curve, one may retrieve them with
 *       a call to 'tpt_fusion_getVariableXAxis()' or 'tpt_fusion_getVariableYAxis()'.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param value
 *          The pointer from which the variable's new value(s) will be copied from. 
 * @param length
 *          The number of values to copy from the pointer to the variable.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeValues(tpt_fusion_node_handle handle, 
                            tpt_fusion_variable_handle vhandle, 
                            const double *values, 
                            unsigned int length);

/**
 * Writes nwe long values to the given variable on the FUSION. If the given variable 
 * is a map or curve, this function will always write the values. The total 
 * number of values copied from the given value pointer depends on the given total "length"
 * of the values. If length > 1, the array is assumed to be 1-dimensional only. Use 
 * tpt_fusion_writeNDimValuesAsLong() to write n-dimensional arrays.
 * 
 * NOTE: In order to write an axis of a map or curve, one may retrieve them with
 *       a call to 'tpt_fusion_getVariableXAxis()' or 'tpt_fusion_getVariableYAxis()'.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param value
 *          The pointer from which the variable's new value(s) will be copied from. 
 * @param length
 *          The number of values to copy from the pointer to the variable.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeValuesAsLong(tpt_fusion_node_handle handle, 
                                  tpt_fusion_variable_handle vhandle, 
                                  const tpt_long *values, 
                                  unsigned int length);

/**
 * Writes the double values of a matrix to the FUSION.
 *
 * The total number of values must be xlength * ylength. After this function has been 
 * used the xlength and the ylength can be read from FUSION
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param value
 *          The pointer from which the variable's new value(s) will be copied from. 
 *          The values must be stored line-by-line, i.e. "index = (x + xlength * y)"
 * @param xlength
 *          The number of values to copy from the pointer to the variable.
 * @param ylength
 *          The number of values to copy from the pointer to the variable.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeMatrixValue(tpt_fusion_node_handle handle, 
                                 tpt_fusion_variable_handle vhandle, 
                                 const double *values, 
                                 unsigned int xlength, 
                                 unsigned int ylength);

/**
 * Writes the long values of a matrix to the FUSION.
 *
 * The total number of values must be xlenght * ylength. After this function has been 
 * used the xlength and the ylength can be read from FUSION
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write. 
 * @param value
 *          The pointer from which the variable's new value(s) will be copied from. 
 *          The values must be stored line-by-line, i.e. "index = (x + xlength * y)"
 * @param xlength
 *          The number of values to copy from the pointer to the variable.
 * @param ylength
 *          The number of values to copy from the pointer to the variable.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeMatrixValueAsLong(tpt_fusion_node_handle handle, 
                                       tpt_fusion_variable_handle vhandle, 
                                       const tpt_long *values, 
                                       unsigned int xlength, 
                                       unsigned int ylength);

/**
 * Writes new double values to the given variable on the FUSION as a multi-dimensional 
 * array. 
 *
 * Assuming an array a[2][4][10]. This array is a 3-dim array with the 3 lengths 2, 4, and
 * 10. The total number of elements (total length) is the product 2*4*10=80. In order to
 * write such an array, just call this function with dim=3 and "lengths=[2,4,10]". It is 
 * assumed that the first length in the length array is the outermost one. This means the
 * index in the flat list of "values" for an element a[i][j][k] is computed as 
 *   index = i*lengths[1]*lengths[2] + lengths[2]*j + k
 * or more specific:
 *   index = i*4*10 + 10*j + k
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write.
 * @param value
 *          The pointer from which the variable's new value(s) will be copied from. 
 * @param dim
 *          The dimensions of the multi-dimensional array. dim = 0 means scalar, dim = 1
 *          is a linear array, dim = 2 is a 2-dim array (such as a matrix) and so on.
 * @param lengths
 *          lengths is an integer array with "dim" elements that specifies the lengths of
 *          the multi-dimensional array w.r.t. all the different dimensions. The total length
 *          of the flat "values" array is the product of all lengths specified here.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeNDimValues(tpt_fusion_node_handle handle, 
                                tpt_fusion_variable_handle vhandle, 
                                const double *values, 
                                unsigned int  dim, 
                                const unsigned int* lengths);

/**
 * Writes new long values to the given variable on the FUSION as a multi-dimensional 
 * array. 
 *
 * Assuming an array a[2][4][10]. This array is a 3-dim array with the 3 lengths 2, 4, and
 * 10. The total number of elements (total length) is the product 2*4*10=80. In order to
 * write such an array, just call this function with dim=3 and "lengths=[2,4,10]". It is 
 * assumed that the first length in the length array is the outermost one. This means the
 * index in the flat list of "values" for an element a[i][j][k] is computed as 
 *   index = i*lengths[1]*lengths[2] + lengths[2]*j + k
 * or more specific:
 *   index = i*4*10 + 10*j + k
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable to write.
 * @param value
 *          The pointer from which the variable's new value(s) will be copied from. 
 * @param dim
 *          The dimensions of the multi-dimensional array. dim = 0 means scalar, dim = 1
 *          is a linear array, dim = 2 is a 2-dim array (such as a matrix) and so on.
 * @param lengths
 *          lengths is an integer array with "dim" elements that specifies the lengths of
 *          the multi-dimensional array w.r.t. all the different dimensions. The total length
 *          of the flat "values" array is the product of all lengths specified here.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_writeNDimValuesAsLong(tpt_fusion_node_handle handle, 
                                      tpt_fusion_variable_handle vhandle, 
                                      const tpt_long *values, 
                                      unsigned int  dim, 
                                      const unsigned int* lengths);

/* =============================== FUNCTIONS ========================================== */

/** 
 * declare a function. If a function with the given name already exists, it will be delivered. 
 * Otherwise, a new function will be created. Each function can be declared as often as necessary
 * (the returned instance is the same).
 */
TPT_DECLARE_FUSION(tpt_fusion_function_handle) tpt_fusion_declareFunction(tpt_fusion_node_handle handle,
                                                      const char *name);

/**
 * Retrieves the FUSION session's function for the given name or returns zero/null 
 * if a function with that name has not been declared before. 
 *
 * @param handle
 *          The handle of the calling node.
 * @param name 
 *          The name of the function to retrieve.
 *
 * @return The FUSION session's function for the given name or returns zero/null 
 *         if a function with that name has not been declared.
 */
TPT_DECLARE_FUSION(tpt_fusion_function_handle) tpt_fusion_getFunction(tpt_fusion_node_handle handle,
                                                  const char *name);
                                                  
/**
 * Returns an array of all functions that have been declared on the FUSION. The 
 * resulting array's length can be retrieved via 'tpt_fusion_getFunctionsCount()'.
 *
 * NOTE: Callers must free the returned array with 'tpt_fusion_freeArray()'.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return An array of all functions that have been declared on the FUSION. 
 */
TPT_DECLARE_FUSION(const tpt_fusion_function_handle *) tpt_fusion_getFunctions(tpt_fusion_node_handle handle);

/**
 * Returns the total count of all functions declared on the FUSION.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return The total count of all functions declared on the FUSION.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getFunctionsCount(tpt_fusion_node_handle handle);


/**
 * install a function. If there is another function handler installed already an error will
 * occur.
 *
 * @param handle
 *          The handle of the calling node.
 * @param fhandle 
 *          The handle of the function.
 * @param handler
 *          The actual function handler (which is a function pointer) that should be 
 *          installed for the function
 * @param args
 *          An arbitrary pointer that can be used by callers to mimic closures: The "args"
 *          pointer will be stored within the FUSION repository and will be delivered to
 *          the "handler" function whenever the function is called.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_installFunctionHandler(tpt_fusion_node_handle handle,
                                      tpt_fusion_function_handle fhandle,
                                      tpt_fusion_function_handler handler,
                                      void *args);

/** 
 * call a function. Note, that a function handler must be installed before calling the
 * function.
 *
 * @param handle
 *          The handle of the calling node.
 * @param fhandle 
 *          The handle of the function.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_callFunction(tpt_fusion_node_handle handle,
                            tpt_fusion_function_handle fhandle);

/**
 * Returns the name of the given function. 
 *
 * @param handle
 *          The handle of the calling node.
 * @param fhandle
 *          The handle of the function. 
 *
 * @return The name of the given function. 
 */
TPT_DECLARE_FUSION(const char *) tpt_fusion_getFunctionName(tpt_fusion_node_handle handle, 
                                       tpt_fusion_function_handle fhandle);

/* =============================== VARIABLE INFOS ========================================== */

/**
 * Returns the name of the given variable. 
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 *
 * @return The name of the given variable. 
 */
TPT_DECLARE_FUSION(const char *) tpt_fusion_getVariableName(tpt_fusion_node_handle handle, 
                                       tpt_fusion_variable_handle vhandle);

/**
 * Returns the length of the given variable, following these three rules:
 *
 * 1. For a scalar variable, always returns '1'.
 * 2. For an array variable (including a string variable), returns the number of its elements.
 * 3. For a map or curve variable, returns the full length of their values.
 *
 * In order to get the length of a map's x- or y-axis, one may retrieve them with
 * a call to 'tpt_fusion_getVariableXAxis()' or 'tpt_fusion_getVariableYAxis()'.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 *
 * @return The length of the given variable. 
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getVariableLength(tpt_fusion_node_handle handle, 
                                          tpt_fusion_variable_handle vhandle);

/**
 * The first dimension of a matrix. If the value has not been written by tpt_fusion_writeMatrixValue
 * the returned value will be 0.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getVariableXLength(tpt_fusion_node_handle handle, 
                                           tpt_fusion_variable_handle vhandle);

/**
 * The second dimension of a matrix. If the value has not been written by tpt_fusion_writeMatrixValue
 * the returned value will be 0.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getVariableYLength(tpt_fusion_node_handle handle, 
                                           tpt_fusion_variable_handle vhandle);

/**
 * Returns the dimension "n" of an n-dimensional array.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 * @param dimVar
 *          reference to a variable that will contain the dimension of the variable after the
 *          function returns.
 * @param lengthsVar
 *          reference to a variable that will contain a read-only pointer to an array with "*dimVar"
 *          elements. Each element represents the length of the array in one particular dimension.
 *          See tpt_fusion_writeNDimValues() and tpt_fusion_writeNDimValuesAsLong() for details.
 *
 * NOTE: the dimension stored in "dimVar" is NOT the total length of the underlying flat values 
 *       array. For [4][7][3] dims is 3, but total length is 4*7*3. So, the total length is equal
 *       to the product of all lengths in the lengthsVar array instead.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_getVariableNDimLength(tpt_fusion_node_handle handle, 
                                      tpt_fusion_variable_handle vhandle,
                                      unsigned int *dimVar,
                                      const unsigned int **lengthsVar);

/**
 * Returns the type of the given variable. See enum type 'tpt_fusion_variable_type'.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 *
 * @return The type of the given variable. 
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_type) tpt_fusion_getVariableType(tpt_fusion_node_handle handle, 
                                                    tpt_fusion_variable_handle vhandle);

/**
 * Returns the role of the given variable. See enum type 'tpt_fusion_variable_role'.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 *
 * @return The role of the given variable. 
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_role) tpt_fusion_getVariableRole(tpt_fusion_node_handle handle, 
                                                    tpt_fusion_variable_handle vhandle);

/**
 * Returns the given variable's x-axis or zero/null, if the variable 
 * is not a map or curve.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 *
 * @return the given variables x-axis or zero/null, if the variable is not a map or 
 *         curve.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_getVariableXAxis(tpt_fusion_node_handle handle, 
                                                       tpt_fusion_variable_handle vhandle);

/**
 * Returns the given variable's y-axis or zero/null, if the variable is not a map.
 *
 * @param handle
 *          The handle of the calling node.
 * @param vhandle
 *          The handle of the variable. 
 *
 * @return the given variable's y-axis or zero/null, if the variable is not a map.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_getVariableYAxis(tpt_fusion_node_handle handle, 
                                                       tpt_fusion_variable_handle vhandle);

/**
 * FUSION tracks parameter changes for each node during simulation time. A reset will be 
 * performed each time the call to a node's 'tpt_fusion_node_callfcn' returned. 
 *
 * Successive calls to this function return all the parameters, that have been changed 
 * since the last time, this node's function 'tpt_fusion_node_callfcn' has been called. 
 *
 * @param handle
 *          The handle of the calling node.
 * @return The handle of a parameter that has ...
 *           1. ... been changed after the last cyclic call of the given node, and ...
 *           2. ... has not yet been delivered by a call to this function.
 *         Returns zero/null, if there is no more changed parameter available.
 */
TPT_DECLARE_FUSION(tpt_fusion_variable_handle) tpt_fusion_getNextChangedParam(tpt_fusion_node_handle handle);

/* =============================== TIMER/CLOCK OPERATIONS ========================================== */

/**
 * Defines the response timeout in seconds for the following functions of the node:
 * - tpt_fusion_node_startup(..)
 * - tpt_fusion_node_initfcn(..)
 * - tpt_fusion_node_callfcn(..)
 * - tpt_fusion_node_shutdown(..)
 *
 * It can be called within each of these functions. It is sufficient to call it once.
 * If this function is not called, the response timeout is 10 seconds.
 * 
 * @param handle
 *          The FUSION.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_setFcnResponseTimeOut(tpt_fusion_node_handle handle, unsigned short seconds);

/**
 * Resets the real time clock. 
 * 
 * @param handle
 *          The FUSION.
 */
TPT_DECLARE_FUSION(void) tpt_fusion_resetRtcSessionTicks(tpt_fusion_node_handle handle);

/**
 * Returns the frequency of the real time clock. 
 * 
 * @param handle
 *          The FUSION.
 */
TPT_DECLARE_FUSION(signed tpt_long) tpt_fusion_getRtcTicksPerSec(tpt_fusion_node_handle handle);

/**
 * Returns the current tick count of the real time clock. 
 * 
 * @param handle
 *          The FUSION.
 */
TPT_DECLARE_FUSION(signed tpt_long) tpt_fusion_getRtcCurrentTicks(tpt_fusion_node_handle handle);

/**
 * Returns the session tick count of the real time clock. 
 * 
 * @param handle
 *          The FUSION.
 */
TPT_DECLARE_FUSION(signed tpt_long) tpt_fusion_getRtcSessionTicks(tpt_fusion_node_handle handle);

/**
 * Returns the step size of the simulation in microseconds.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return the basic step size used by the FUSION session
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getStepSize(tpt_fusion_node_handle handle);

/**
 * Returns the elapsed number of simulation steps.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return the elapsed number of simulation steps.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_getCycleCount(tpt_fusion_node_handle handle);

/* =============================== MISC ========================================== */

/**
 * Returns whether FUSION is in compile run.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return FUSION is in compile run.
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_isCompileRun(tpt_fusion_node_handle handle);

/**
 * Returns the current error code from FUSION.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @return FUSION error code.
 */
TPT_DECLARE_FUSION(tpt_fusion_errcode) tpt_fusion_getErrorCode(tpt_fusion_node_handle handle);

/**
 * Returns whether FUSION can communicate with TPT and the specified plugin is licensed in TPT
 *
 * @param pluginName
 *          The name of the plugin
 *
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_isPluginLicensed(const char *pluginName);

/**
 * Formats a message and sends it to standard output.
 *
 * @param handle
 *          The handle of the calling node.
 *
 * @param format
 *          The format of the message.
 *
 * @param ...
 *          An arbitrary number of arguments to the message format.
 *
 */
TPT_DECLARE_FUSION(void) tpt_fusion_printf(tpt_fusion_node_handle handle, const char *format, ...);

/**
 * Frees the given array, which has been returned by a call to FUSION.
 *
 * @param farray
 *          The address of the array to be freed.
 *
 */
TPT_DECLARE_FUSION(void) tpt_fusion_freeArray(void *address);


/**
 * enqueue the job "job" into the asynchronous execution thread. This function can be used to enqueue processing functions
 * that last too long for a regular simulation cycle and should be handled in parallel. Note, that there is just *one, single*
 * executor thread handling all jobs. Jobs will be enqueued and processed in FIFO order.
 * 
 * @param handle
 *          The handle of the calling node.
 *
 * @param job
 *          Function pointer to the acutal job to process.
 *
 * @param data
 *          Pointer to some data structure that should be hand over to the job as soon as exeuction starts.
 * 
 * @return A handle to the enqueued job. This handle can be used to check the status of the job using 
 *          tpt_fusion_isRunAsyncReady() later.
 */
TPT_DECLARE_FUSION(tpt_fusion_job_handle) tpt_fusion_runAsync(tpt_fusion_node_handle handle, tpt_fusion_job job, void *data);

/**
 * check if a job has been processed completely. Note that this information is volatile and can be checked just once. 
 * So if the function returned true once and is called repeatedly, it returns false again.
 * 
 * @param handle
 *          The handle of the calling node.
 *
 * @param jobHandle
 *          handle to the enqueued job.
 * 
 * @return true if the job has been processed completely. Note that a "true"-flag is volatile and is returned just once!
 */
TPT_DECLARE_FUSION(unsigned int) tpt_fusion_isRunAsyncReady(tpt_fusion_node_handle handle, tpt_fusion_job_handle jobHandle);

#ifdef __cplusplus
}
#endif

#endif /* __TPT_FUSION_H_ */
