/*********************************************************************
 *              Copyright (c) 2021 PikeTec GmbH
 *               CONFIDENTIAL AND PROPRIETARY.
 *                   ALL RIGHTS RESERVED
 *              - PROPRIETARY OF PikeTec GmbH -
 **********************************************************************/
/* clang-format off */

/*
 * defines core data types of the TPT virtual machine. 
 * 
 * PikeTec tries to keep this interface as compatible as possible between TPT 
 * versions without causing users too much pain when migrating from one version to the next. 
 * Nevertheless, compatibility cannot always be 100% guaranteed. Therefore, all users of 
 * this API are advised to check compatibility and adjust if necessary before switching to 
 * a new version.
 */

#ifndef __TPT_TYPES_H_
#define __TPT_TYPES_H_

/* ------------------------- */



typedef double         tpt_double;    /* double = 64 bit floating point (IEEE 754) */
typedef unsigned char  tpt_bool;      /* bool = boolean type (represented as 8 bit) */
typedef float          tpt_float;     /* float = 32 bit floating point (IEEE 754) */
typedef unsigned char  tpt_uint8;     /* uint8 = 8 bit unsigned integer */
typedef signed char    tpt_int8;      /* int8 = 8 bit signed integer */
typedef unsigned short tpt_uint16;    /* uint16 = 16 bit unsigned integer */
typedef signed short   tpt_int16;     /* int16 = 16 bit signed integer */
typedef unsigned int   tpt_uint32;    /* uint32 = 32 bit unsigned integer */
typedef signed int     tpt_int32;     /* int32 = 32 bit signed integer */
typedef tpt_uint8      tpt_byte;      /* single byte (unsigned) */
typedef tpt_uint32     tpt_tsize;     /* data type for array indices and array dimensions (same as uint32) */

#if defined(_MSVC_LANG)
  typedef __int64      tpt_int64;     /* int64 = 64 bit signed integer (MSVC)  */
#else  
  typedef signed long long tpt_int64; /* int64 = 64 bit signed integer (GCC) */
#endif

#define tpt_true   ((tpt_bool)1)      /* constant true (same as byte 0x01) */
#define tpt_false  ((tpt_bool)0)      /* constant false (same as byte 0x00) */
#define TPT_NULL   0                  /* general purpose 0 or nullptr */

/* enumeration of all primitive data types */
typedef enum {
    tpt_type_uint8      = 1,  /* uint8 = 8 bit unsigned integer */
    tpt_type_int8       = 2,  /* int8 = 8 bit signed integer */
    tpt_type_uint16     = 3,  /* uint16 = 16 bit unsigned integer */
    tpt_type_int16      = 4,  /* int16 = 16 bit signed integer */
    tpt_type_uint32     = 5,  /* uint32 = 32 bit unsigned integer */
    tpt_type_int32      = 6,  /* int32 = 32 bit signed integer */
    tpt_type_int64      = 7,  /* int64 = 64 bit signed integer (GCC) */
    tpt_type_float      = 8,  /* float = 32 bit floating point (IEEE 754) */
    tpt_type_double     = 9,  /* double = 64 bit floating point (IEEE 754) */
    tpt_type_bool       = 10, /* bool = boolean type (represented as 8 bit) */

    tpt_type_any        = -1  /* special const representing an "irrelevant" basetype (similar to a "nullptr").
                               * This is **not** a regular data type. It is used just for some API functions
                               * to represent a "not-relevant". Whereever this value is accepted, it is explicitly
                               * mentioned. */
} tpt_basetype;


/* enumeration of general purpose return codes. */
typedef enum {
  tpt_result_okay       = 0,  /* used as return value of functions if the function terminates regularily (no error) */
  tpt_result_error      = 1,  /* used as return value of functions if the function terminates with an exception/error */
  tpt_result_terminated = 2,  /* special-purpose return code used by tpt_runCycle() only (see tpt_runCycle() for details). */
} tpt_result;


/* define calling convention */
#if !defined(TPT_DECLARE)
  #if defined(_TPTDLL) /* <- not for user purpose */
    #define TPT_DECLARE(x) __declspec(dllexport) x
  #elif defined(TPT_VMAPI_STATIC)
    #define TPT_DECLARE(x) x
  #elif defined(__LCC__)
    #define TPT_DECLARE(x) x
  #else
    #define TPT_DECLARE(x) __declspec(dllimport) x
  #endif
#endif

#endif /* __TPT_TYPES_H_ */
