package com.etas.ascet.api;

import java.util.List;

import org.eclipse.jdt.annotation.NonNullByDefault;

import com.etas.ascet.api.v1.AscetApiV1;
import com.etas.ascet.api.v1.ClientApiV1;
import com.google.auto.value.AutoValue;
import com.ryanharter.auto.value.gson.GenerateTypeAdapter;

// Checkstyle:off
@NonNullByDefault
@GenerateTypeAdapter
@AutoValue
public abstract class AscetDetailsDto {

    /**
     * The product name, e.g. "ASCET-DEVELOPER"
     */
    public abstract String name();

    /**
     * The supported API versions. The numbers must not be consecutive.<br/>
     * The client uses this to chose a version that the client supports as well and then used the matching interfaces.
     * E.g. to use version 1, the interfaces are {@link AscetApiV1} and {@link ClientApiV1}.<br/>
     * Note: an API might be "in development", then it has {@link ApiVersionDto#isReleased()} configured as false.
     */
    public abstract List<ApiVersionDto> supportedApiVersions();

    /**
     * The ASCET-DEVELOPER version in format {@code <major>.<minor>.<fix>}, e.g. "7.8.0"<br/>
     * This string is for information only and shall not be parsed. Instead the separated fields shall be used.
     */
    public abstract String versionName();

    /**
     * Version in format {@code <major>.<minor>.<fix>[-<preRelease>][-<build>]}.<br/>
     * This string is for information only and shall not be parsed. Instead the separated fields shall be used.<br/>
     * Examples:
     * <ul>
     * <li>{@code "7.8.0"}
     * <li>{@code "7.8.0-#34"}
     * <li>{@code "7.8.0-BETA-#34"}
     * </ul>
     */
    public abstract String versionFullName();

    /**
     * The major (first) version number. A positive integer or zero.
     */
    public abstract int majorVersion();

    /**
     * The minor (middle) version number. A positive integer or zero.
     */
    public abstract int minorVersion();

    /**
     * The fix (last) version number. A positive integer or zero.
     */
    public abstract int fixVersion();

    /**
     * Either empty for release version or a qualifier like e.g. "BETA".
     * If {@link #preRelease()} is empty, then {@link #isReleased()} is true.
     */
    public abstract String preRelease();

    /**
     * Either empty string or a qualifier that represents the build number. The value can be non-numerical. E.g. "#34"
     */
    public abstract String buildNumber();

    /**
     * {@code true} for release version. A released version does have an empty string for {@link #preRelease()}.
     */
    public abstract boolean isReleased();

    /**
     * Capabilities that are supported by the running instance. These are features that might be installed optionally.
     * E.g. a customer specific extension.
     */
    public abstract List<String> capabilities();

    /**
     * Absolute file path of the ASCET installation
     */
    public static final String PROPERTY_KEY_INSTALLATION_PATH = "installation.path";
    /**
     * Absolute file path of currently used workspace
     */
    public static final String PROPERTY_KEY_WORKSPACE_PATH = "workspace.path";
    /**
     * OS specific process ID of the running ASCET instance.
     * This can be used to forcibly kill the ASCET-DEVELOPE application or to track if the process is alive.
     */
    public static final String PROPERTY_KEY_PROCESS_ID = "process.id";

    /**
     * List of key/value pairs. The keys are added as string constants in this interface.
     */
    public abstract List<KeyValueDto> properties();

    public static Builder builder() {
        return new AutoValue_AscetDetailsDto.Builder();
    }

    @AutoValue.Builder
    public interface Builder {
        /**
         * @see AscetDetailsDto#name()
         */
        Builder name(String value);

        /**
         * @see AscetDetailsDto#supportedApiVersions()
         */
        Builder supportedApiVersions(List<ApiVersionDto> value);

        /**
         * @see AscetDetailsDto#versionName()
         */
        Builder versionName(String value);

        /**
         * @see AscetDetailsDto#versionFullName()
         */
        Builder versionFullName(String value);

        /**
         * @see AscetDetailsDto#major()
         */
        Builder majorVersion(int value);

        /**
         * @see AscetDetailsDto#minor()
         */
        Builder minorVersion(int value);

        /**
         * @see AscetDetailsDto#fix()
         */
        Builder fixVersion(int value);

        /**
         * @see AscetDetailsDto#preRelease()
         */
        Builder preRelease(String value);

        /**
         * @see AscetDetailsDto#buildNumber()
         */
        Builder buildNumber(String value);

        /**
         * @see AscetDetailsDto#isReleased()
         */
        Builder isReleased(boolean value);

        /**
         * @see AscetDetailsDto#capabilities()
         */
        Builder capabilities(List<String> value);

        /**
         * @see AscetDetailsDto#properties()
         */
        Builder properties(List<KeyValueDto> value);

        AscetDetailsDto build();
    }
}
