package com.etas.ascet.api.protocol;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;
import org.eclipse.lsp4j.jsonrpc.Launcher;
import org.eclipse.lsp4j.jsonrpc.RemoteEndpoint;

import com.ryanharter.auto.value.gson.GenerateTypeAdapter;

/**
 * Configures the JSON-RPC {@link Launcher} to be used with network socket and ASCET API
 *
 * @param <T>
 *            The remote side interface. If more than one is given, then it will be Object and must be casted as needed.
 */
@NonNullByDefault
public class SocketLauncher<T> implements Launcher<T> {

    private final Launcher<T> launcher;
    @Nullable
    private Future<Void> startListeningFuture;
    private final ExecutorService executorService;
    private final Closeable connection;

    /**
     * @param localServices
     *            all implementation of the local object that can be called by the remote side
     * @param remoteInterfaces
     *            all interfaces that shall be proxied. Because they are combined into a single remote proxy object,
     *            those interfaces cannot contains conflicting method signatures. E.g. same method name and argument
     *            types with different return type cannot co-exist in the proxy object.
     * @param connection
     *            a closable (e.g. the socket) that will be closed when listening is cancelled.
     * @param inputStream
     *            the input stream of the communication channel, e.g. <code>socket.getInputStream()</code>
     * @param outputStream
     *            the output stream of the communication channel, e.g. <code>socket.getOutputStream()</code>
     * @param executorService
     *            the thread pool to be used to run the activities on.
     */
    public SocketLauncher(final List<Object> localServices,
            final List<Class<? extends @Nullable T>> remoteInterfaces,
            final Closeable connection,
            final InputStream inputStream,
            final OutputStream outputStream,
            final ExecutorService executorService) {
        this.connection = connection;
        this.executorService = executorService;
        try {
            this.launcher = new Launcher.Builder<T>()
                    .setLocalServices(localServices)
                    .setRemoteInterfaces(remoteInterfaces)
                    .setInput(inputStream)
                    .setOutput(outputStream)
                    .setClassLoader(getClass().getClassLoader())
                    .setExecutorService(executorService)
                    .configureGson(gsonBuilder -> gsonBuilder
                            .registerTypeAdapterFactory(GenerateTypeAdapter.FACTORY)
                            .registerTypeAdapterFactory(OptionalTypeAdapter.FACTORY))
                    .create();
        } catch (final Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * @param localServices
     *            all implementation of the local object that can be called by the remote side
     * @param remoteInterfaces
     *            all interfaces that shall be proxied. Because they are combined into a single remote proxy object,
     *            those interfaces cannot contains conflicting method signatures. E.g. same method name and argument
     *            types with different return type cannot co-exist in the proxy object.
     * @param socket
     *            the socket for communication
     * @param executorService
     *            the thread pool to be used to run the activities on.
     * @throws IOException
     */
    public SocketLauncher(final List<Object> localServices,
            final List<Class<? extends @Nullable T>> remoteInterfaces,
            final Socket socket,
            final ExecutorService executorService) throws IOException {
        this(localServices, remoteInterfaces, socket,
                socket.getInputStream(), socket.getOutputStream(), executorService);
    }

    @NonNullByDefault({})
    @Override
    public CompletableFuture<Void> startListening() {
        final var startListeningFuture2 = this.launcher.startListening();
        startListeningFuture = startListeningFuture2;

        final CompletableFuture<Void> result = CompletableFuture.runAsync(() -> {
            try {
                startListeningFuture2.get();
            } catch (final Exception e) {
                throw new RuntimeException(e);
            }
        }, executorService);

        return result;
    }

    /**
     * Stop the server listening
     *
     * @param mayInterruptIfRunning
     */
    public void cancel(final boolean mayInterruptIfRunning) {
        // first do the cancel, afterwards the connection.close()
        // the cancel does inform that we want to terminate,
        // then the connection.close() is not causing a logging about "socket closed"
        final var startListeningFuture2 = startListeningFuture;
        if (startListeningFuture2 != null) {
            startListeningFuture2.cancel(mayInterruptIfRunning);
        }
        try {
            connection.close();
        } catch (final IOException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public T getRemoteProxy() {
        return this.launcher.getRemoteProxy();
    }

    @Override
    public RemoteEndpoint getRemoteEndpoint() {
        return this.launcher.getRemoteEndpoint();
    }
}
